//
// markIV: software to download CCD data from a MarkIV ISA card and parse
//         two files containing FITS style header information, with the intent
//         of creating two FITS files containing appropriate information
// Copyright (C) 2001  Robert S. Creager
//
// E-mail: robert@carolinecreager.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// $Name: Downfits-1_1 $
//
// $Log: tom.h,v $
// Revision 1.0  2002/01/06 18:12:20  robc
// First CVS entry
//                                                                         

#ifndef _TOM_H
#define _TOM_H

#include <fstream.h>
#include "splash.h"
#include "splistit.h"

#define MAX_CCD 2

// The core function, which churns through the given input file and parses
// it looking for FITS style header information to add to the given fits file
extern void parse_main( fitsfile*[ MAX_CCD ], char[ MAX_CCD ], int );
extern void add_keywords( fitsfile*, ifstream& );
extern char* char_reserved[];

// List of reserved keywords to be filled out within the program using
// below data
extern const int num_reserved;
extern SPStringList reserved;

#endif
