/*
 * The first byte of the regexp internal "program" is actually this magic
 * number; the start node begins in the second byte.
 */
//
// $Name: Downfits-1_1 $
//
// $Log: regmagic.h,v $
// Revision 1.0  2002/01/06 18:12:19  robc
// First CVS entry
//                                                                         

#define	MAGIC	0234
